/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/errno.h>
#include <sys/buf.h>
#include <sys/malloc.h>
#include <sys/ioctl.h>
#include <sys/device.h>
#include <sys/conf.h>
#include <sys/lock.h>
#include <sys/queue.h>
#include <sys/vnode.h>
#include <sys/fcntl.h>
#include <sys/namei.h>
#include <sys/security.h>
#include <sys/vexec.h>

/* Autoconf. */
void vexecattach(struct device *parent, struct device *self, void *aux) {
	vexec_dprintf(("vexecattach: Vexec pseudo-device attached.\n"));
}

/* Pseudo-open routine. */
int vexecopen(dev_t dev, int flags, int fmt, struct proc *p) {
	if (security_vexec_verbose)
		printf("vexecopen: Vexec load device opened by uid=%u, "
		    "pid=%u. (dev=%d)\n", p->p_ucred->cr_uid, p->p_pid, dev);

	return (0);
}

/* Pseudo-close routine. If verbose, print all entries in the tables. */
int vexecclose(dev_t dev, int flags, int fmt, struct proc *p) {
	if (security_vexec_verbose) {
		struct vexec_hashtbl *tbl;

		LIST_FOREACH(tbl, &vexec_tables, hash_list) {
			printf("vexecclose: dev=%d, inodes=%u\n",
			    tbl->hash_dev, tbl->hash_size);
		}

#if 0
		/* Print entries in all hash tables. */
		printf("vexecclose: Hash table content:\n");
		vexec_tblwalk((VEXEC_CALLBACK *) &vexec_hashprint);
#endif
	}

	return (0);
}

/* Handle ioctls for the device. */
int vexecioctl(dev_t dev, u_long cmd, caddr_t data, int flags, struct proc *p) {
	struct vexec_hashtbl *tbl;
	struct nameidata nid;
	struct vattr va;
	int error = 0;

	vexec_dprintf(("vexec_ioctl: cmd=0x%lx, file=%s\n", cmd,
	    params->vxp_file));

	/*
	 * Don't allow updates in multi-user mode. Give a hint to
	 * the user. ;)
	 */
	if (securelevel >= 1) {
		printf("vexecioctl: Can't load in multi-user mode! "
		    "Try \"kill 1\"...\n");

		return (EPERM);
	}
	
	switch (cmd) {
	case VEXEC_PRELOAD: {
		struct vexec_hash_param *params = (struct vexec_hash_param *) data;
		u_long hashmask;

		/* Allocate and initialize a Vexec hash table. */
		tbl = malloc(sizeof(struct vexec_hashtbl), M_TEMP, M_WAITOK);
		tbl->hash_size = params->hp_size;
		tbl->hash_dev = params->hp_dev;
		tbl->hash_tbl = hashinit(params->hp_size, M_TEMP,
		    M_WAITOK, &hashmask);

		LIST_INSERT_HEAD(&vexec_tables, tbl, hash_list);

		break;
		}

	case VEXEC_LOAD: {
		struct vexec_params *params = (struct vexec_params *) data;
		struct vexec_hash_entry *hh;
		struct vexec_hash_entry *e;

		NDINIT(&nid, LOOKUP, FOLLOW, UIO_SYSSPACE, params->vxp_file, p);
		error = namei(&nid);
		if (error)
			return (error);

		/* Add only regular files. */
		if (nid.ni_vp->v_type != VREG) {
			printf("vexecioctl: Not adding \"%s\": Not a "
			    "regular file.\n", params->vxp_file);
			vrele(nid.ni_vp);
			return (EINVAL);
		}

		/*
		 * Make sure we evaluate fingerprint it in the first access.
		 */
		nid.ni_vp->v_fpstatus = FP_NOTEVAL;

		/* Get attributes for device and inode. */
		error = VOP_GETATTR(nid.ni_vp, &va, p->p_ucred, p);
		if (error)
			return (error);

		/* Release our reference to the vnode. (namei) */
		vrele(nid.ni_vp);

		/* Get table for the device. */
		tbl = vexec_tblfind(va.va_fsid);
		if (tbl == NULL) {
			return (EINVAL);
		}

		/*
		 * HERE WE DO THE CONVERSION BETWEEN params AND A NORMAL
		 * ENTRY IN THE HASH TABLE. THIS EFFECITVELY REPLACED
		 * vexec_inodeadd()'s work.
		 */
		hh = vexec_lookup(va.va_fsid, va.va_fileid);
		if (hh != NULL) {
			/*
			 * Already have entry. This can happen for various
			 * reasons we don't give a fuck about. Deny
			 * modification and yell.
			 */
			if (security_vexec_verbose) {
				printf("vexecioctl: Duplicate entry for "
				    "%s. Ignoring.\n", params->vxp_file);
			}

			return (0);
		}

		/* Create vexec_hash_entry from vexec_params */
		e = malloc(sizeof(*e), M_TEMP, M_WAITOK);
		e->vhe_inode = va.va_fileid;
		e->vhe_type = params->vxp_type;
		e->vhe_fptype = params->vxp_fptype;
		memcpy(e->vhe_fp, params->vxp_fp, sizeof(e->vhe_fp));

		vexec_dprintf(("Vexec: New entry (file=%s, inode=%u, dev=%u)\n",
		    params->vxp_file, va.va_fileid, va.va_fsid));

		error = vexec_hashadd(tbl, e);

		break;
		}

	default:
		/* Invalid operation. */
		error = ENODEV;

		break;
	}

	return (error);
}
