/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/sysctl.h>
#include <sys/vnode.h>
#include <sys/security.h>
#include <sys/vexec.h>

#include <sys/mount.h>
#include <sys/syscallargs.h>

int64_t security_trust_gid = TRUST_INVAL;
int security_tpe = 1;
int security_privacy = 0; /* bitmap */
int security_trust_override = 0;

/*
 * Verify trust of a (user running a) process.
 *
 * The gate for all trust related operations. The code should be able to
 * decide whether the user is trusted or not according to the current
 * trust group-id, and set the status on the process accordingly.
 */
int trusted(struct proc *p) {
	struct pcred *cred = p->p_cred;

	/*
	 * Compare the trusted-group-id in the process with the current
	 * one on the system. If they match, it's trusted.
	 */
	if (security_trust_gid == TRUST_INVAL) {
		cred->p_trustgid = cred->p_rgid;
		return (1);
	} else {
		if (cred->p_trustgid == security_trust_gid)
			return (1);
	}

	/*
	 * Next check is for superuser. The superuser is always trusted.
	 */
	if (suser(p, 0) == 0) {
		cred->p_trustgid = security_trust_gid;
		return (1);
	}

	/*
	 * The last thing to check for is if the user is in the current
	 * trust group. this is done with a call to groupmember().
	 */
	if (groupmember(security_trust_gid, p->p_ucred)) {
		cred->p_trustgid = security_trust_gid;
		return (1);
	}

	/*
	 * If we're here it means we're not trusted. Set the 
	 * trusted-group-id in the process to something invalid.
	 */
	cred->p_trustgid = TRUST_INVAL;

	return (0);
}

/*
 * Privacy: Check if "looker" can get information on "object".
 */
int proc_cansee(struct proc *looker, struct proc *object) {
	/*
	 * We allow "looker" to access "object" if either of these is true:
	 *
	 *  - Process privacy is disabled
	 *  - They have the same user-id
	 *  - "looker" is trusted and "trust overrides privacy"
	 */
	if (!security_privacy ||
	    (looker->p_ucred->cr_uid == object->p_ucred->cr_uid) ||
	    (suser(looker, 0) == 0) ||
	    (security_trust_override && trusted(looker)))
		return (1);

	return (0);
}

/*
 * Privacy: check if the specified component has privacy enabled or not.
 */
int private(struct proc *p, int component) {
	switch (component) {
	case KERN_PRIVACY_OP:
	case KERN_PRIVACY_PROCFS:
	case KERN_PRIVACY_NETSTAT:
	case KERN_PRIVACY_W:
	case KERN_PRIVACY_WHO:
	case KERN_PRIVACY_LAST:
	case KERN_PRIVACY_FINGER:
		break;

	default:
		return (0);
	}

	if (BITMAP_ISSET(security_privacy, component))
		return (1);
	else
		return (0);
}

/*
 * Handle kernel security sysctls.
 * Used for both checking trust of a process and status of a feature.
 */
int
sysctl_security(name, namelen, oldp, oldlenp, newp, newlen, p)
	int *name;
	u_int namelen;
	void *oldp;
	size_t *oldlenp;
	void *newp;
	size_t newlen;
	struct proc *p;
{
	int64_t tmp_trust_gid;
	int r, security_privacy_pass = 0;

	if ((namelen != 2) &&
	    (name[0] != KERN_SECURITY_VEXEC) &&
	    (name[0] != KERN_SECURITY_PRIVACY))
		return (ENOTDIR);		

	switch (name[0]) {
	case KERN_SECURITY_TRUSTED: {
		int trust;

		if (security_trust_gid == TRUST_INVAL)
			trust = (suser(p, 0) == 0) ? 1 : 0;
		else
			trust = trusted(p);

		return (sysctl_int(oldp, oldlenp, NULL, 0, &trust));
		}

	case KERN_SECURITY_TRUSTGID:
		/*
		 * There's no point in disallowing modification of the
		 * trust group-id in the highly secure mode(s) because it
		 * can be more dangerous than helpful.
		 *
		 * It's safe to change the trust group-id anytime because
		 * we take caution to avoid race conditions -- read more
		 * in the documentation or sys/vexec.h.
		 */
		tmp_trust_gid = security_trust_gid;
		r = sysctl_quad(oldp, oldlenp, newp, newlen,
		    &tmp_trust_gid);
		if (r)
			return (r);

		if (newp && !(((tmp_trust_gid >= 0) &&
		    (tmp_trust_gid <= GID_MAX)) ||
		    (tmp_trust_gid == TRUST_INVAL)))
			return (EINVAL);

		security_trust_gid = tmp_trust_gid;

		return (r);

	case KERN_SECURITY_TRUST_OVERRIDE:
		return (sysctl_int(oldp, oldlenp, newp, newlen, 
		    &security_trust_override));

	case KERN_SECURITY_TPE:
		return (sysctl_int(oldp, oldlenp, newp, newlen,
		    &security_tpe));

	case KERN_SECURITY_VEXEC:
		switch (name[1]) {
		case KERN_VEXEC_OP:
			/*
			 * In multi-user mode(s), if Vexec is already
			 * enabled just return because it's boolean and we
			 * don't want to lower it.
			 */
			if ((securelevel > 0) && security_vexec && newp)
				return (EPERM);

			return (sysctl_int(oldp, oldlenp, newp, newlen, 
			    &security_vexec));

		case KERN_VEXEC_VERBOSE:
			return (sysctl_int(oldp, oldlenp, newp, newlen,
			    &security_vexec_verbose));

		case KERN_VEXEC_STRICT:
			/* Raise only in multi-user. */
			if ((securelevel > 0) && security_vexec_strict && newp)
				return (EPERM);

			return (sysctl_int(oldp, oldlenp, newp, newlen,
			    &security_vexec_strict));

		default:
			return (EOPNOTSUPP);
		}

	case KERN_SECURITY_PRIVACY:
		/*
		 * Privacy flags are stored in a bitmap. Provide
		 * user-transparent translation between the on/off bit as
		 * a fictive sysctl variable.
		 */
		switch (name[1]) {
		case KERN_PRIVACY_OP:
		case KERN_PRIVACY_PROCFS:
		case KERN_PRIVACY_NETSTAT:
		case KERN_PRIVACY_W:
		case KERN_PRIVACY_WHO:
		case KERN_PRIVACY_LAST:
		case KERN_PRIVACY_FINGER:
			break;

		default:
			return (EOPNOTSUPP);
		}

		/* Get current state */
		security_privacy_pass = BITMAP_ISSET(security_privacy,
		    (int) name[1]) ? 1 : 0;

		r = sysctl_int(oldp, oldlenp, newp, newlen,
		    &security_privacy_pass);
		if (r)
			return (r);

		if (security_privacy_pass)
			BITMAP_SET(security_privacy, (int) name[1]);
		else
			BITMAP_CLR(security_privacy, (int) name[1]);

		return (r);

	default:
		return (EOPNOTSUPP);
	}

	/* NOTREACHED */
}
