/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/mount.h> 
#include <sys/malloc.h>
#include <sys/vnode.h>
#include <sys/namei.h>
#include <sys/exec.h>
#include <sys/proc.h>
#include <sys/security.h>
#include <sys/vexec.h>

/* Pulled from sysctl */
int security_vexec = 0;
int security_vexec_verbose = 0;
int security_vexec_strict = 0;

/* Vexecs table of hash types and their associated information. */
struct vexec_hash vexec_hash_list[] = {
    { VEXEC_HASHINIT(MD5) },
    { VEXEC_HASHINIT(SHA1) },
    { VEXEC_HASHINIT2(SHA256) },
    { VEXEC_HASHINIT2(SHA384) },
    { VEXEC_HASHINIT2(SHA512) },
    { VEXEC_HASHINIT(RMD160) }
};

/* Read buffer for fingerprint calculation. */
#define	VEXEC_BUFSIZE	PAGE_SIZE

/*
 * Calculate the fingerprint of a given file.
 *
 * The fingerprint is calculated according to the hash type.
 */
int vexec_fp_calc(struct proc *p, struct vnode *vp,
    struct vexec_hash_entry *vhe, u_quad_t size, u_char *fp) {
	void *ctx = NULL;
	u_char *buf = NULL;
	u_quad_t offset, len;
	size_t ctxsize, resid;
	int error;

	/* This should not happen. */
	if (!FPTYPE_OK(vhe->vhe_fptype)) {
		panic("vexec_fp_calc: Invalid fingerprint type: %u. "
		    "Report a bug.", vhe->vhe_fptype);
	}

	bzero(fp, FPLEN_MAX);

	/* Get the context size. */
	ctxsize = (vexec_hash_list[vhe->vhe_fptype]).vxh_ctxsize;

	/* Allocate memory for hash context and read buffer. */
	ctx = (void *) malloc(ctxsize, M_TEMP, M_WAITOK);
	buf = (u_char *) malloc(VEXEC_BUFSIZE, M_TEMP, M_WAITOK);
	
	((vexec_hash_list[vhe->vhe_fptype]).vxh_init)(ctx);

	/*
	 * It is assumed that the vnode is locked. execve() does that for
	 * us and we have our own locking in vn_open().
	 */
	for (offset = 0; offset < size; offset += VEXEC_BUFSIZE) {
		len = (offset + VEXEC_BUFSIZE > size) ? (size - offset)
						      : VEXEC_BUFSIZE;

		error = vn_rdwr(UIO_READ, vp, buf, len, offset, 
				UIO_SYSSPACE, 0, p->p_ucred, &resid, p);

		if (error)
			goto bad;

		((vexec_hash_list[vhe->vhe_fptype]).vxh_update)(ctx, buf,
		    (unsigned int) len);
	}

	((vexec_hash_list[vhe->vhe_fptype]).vxh_final)(fp, ctx);

bad:
	free(ctx, M_TEMP);
	free(buf, M_TEMP);

	return (error);
}
	
/*
 * Compare two fingerprints (which are assumed, and this is very important,
 * to be of the same hash type!).
 */
int vexec_fp_cmp(struct vexec_hash_entry *vhe, u_char *digest) {
	return (memcmp(vhe->vhe_fp, digest,
	    (vexec_hash_list[vhe->vhe_type]).vxh_len));
}

/*
 * Get the hash table for the specified device.
 * XXX: Maybe later this could become a hash table aswell?
 */
struct vexec_hashtbl *vexec_tblfind(dev_t device) {
	struct vexec_hashtbl *tbl;

	LIST_FOREACH(tbl, &vexec_tables, hash_list) {
		if (tbl->hash_dev == device)
			return (tbl);
	}

	return (NULL);
}

/* Perform a lookup on a hash table. */
struct vexec_hash_entry *vexec_lookup(dev_t device, ino_t inode) {
	struct vexec_hashtbl *tbl;
	struct vexec_hashhead *tble;
	struct vexec_hash_entry *e;
	size_t indx;

	tbl = vexec_tblfind(device);
	if (tbl == NULL)
		return (NULL);

	indx = VEXEC_HASH(tbl, inode);
	tble = &(tbl->hash_tbl[indx & VEXEC_HASH_MASK(tbl)]);

	LIST_FOREACH(e, tble, vhe_list) {
		/* XXX WORK THIS OUT! XXX */
		if ((e != NULL) && (e->vhe_inode == inode))
			return (e);
	}

	return (NULL);
}

/*
 * Add an entry to a hash table. If a collision is found, handle it.
 * The passed entry is allocated in kernel memory.
 */
int vexec_hashadd(struct vexec_hashtbl *tbl, struct vexec_hash_entry *e) {
	struct vexec_hashhead *vhh;
	size_t indx;

	if (tbl == NULL)
		return (EFAULT);

	indx = VEXEC_HASH(tbl, e->vhe_inode);
	vhh = &(tbl->hash_tbl[indx]);

	if (vhh == NULL)
		panic("vexec_hashadd: vhh is NULL.");

	/* Put in the hash table. */
	LIST_INSERT_HEAD(vhh, e, vhe_list);

	return (0);
}

/*
 * Print a hash table entry. Debugging..
 */
void vexec_hashprint(struct vexec_hash_entry *vhe, dev_t device) {
	if (vhe == NULL)
		return;

	printf("dev=%d, inode=%u, type=%s\n", device, vhe->vhe_inode,
	    (vhe->vhe_type == VEXEC_DIRECT) ? "direct" :
	    (vhe->vhe_type == VEXEC_INDIRECT) ? "indirect" :
	    (vhe->vhe_type == VEXEC_SHELL) ? "shell" :
	    (vhe->vhe_type == VEXEC_FILE) ? "file" :
	    "unknown");
}

/*
 * Hash table traverse. Return each entry to the callback and continue.
 */
int vexec_tblwalk(VEXEC_CALLBACK *callback) {
#if 0
	struct vexec_hashtbl *tbl = NULL;
	struct vexec_hashhead *vhead = NULL;
	struct vexec_hash_entry *vhe = NULL;
	size_t i;
#endif

	if (callback == NULL)
		return (EFAULT);

#if 0
	LIST_FOREACH(tbl, &vexec_tables, hash_list) {
		printf("vexec_tblwalk: looping tables.\n");
		for (i = 0; i < tbl->hash_size; i++) {
			vhead = &(tbl->hash_tbl[i]);
			if (vhead) {
				LIST_FOREACH(vhe, vhead, vhe_list) {
					(*callback)(vhe, tbl->hash_dev);
				}
			}
		}
	}
#endif

	return (0);
}

/*
 * Verify the fingerprint of the given file.
 *
 * 'flag' indicates the execution type we need to enforce, if vexec or TPE
 * is enabled.
 */
int vexec_verify(struct proc *p, struct vnode *vp, struct vattr *va,
    u_char *name, int flag) {
	struct vexec_hash_entry *vhe;
        char digest[FPLEN_MAX];
        int error = 0;

	/* Lie and say it's verified if neither vexec or TPE are enabled. */
	if (!security_vexec && !security_tpe)
		return (0);

	/*
	 * Evaluate fingerprint if needed and set the status on the vp.
	 */
	if (vp->v_fpstatus == FP_NOTEVAL) {
		vhe = vexec_lookup(va->va_fsid, va->va_fileid);
		if (vhe == NULL) {
			vp->v_fpstatus = FP_NOENTRY;
			goto out;
		}
 
		vexec_dprintf(("vexec_verify: Got entry for %s "
		    "(dev=%d, inode=%u)\n", name, va->va_fsid,
		    va->va_fileid));

		/*
		 * If Vexec is not enabled, we only care about the
		 * entry type, to enforce direct/indirect execution
		 * so we prevent interpreters from bypassing TPE.
		 */
		if (!security_vexec) {
			/*
			 * This code is in vexec, but it is used
			 * only for TPE, which is why we're not
			 * modifying the fingerprint status.
			 */
			if (!trusted(p) &&
			    (flag == VEXEC_INDIRECT) &&
			    (vhe->vhe_type != VEXEC_INDIRECT)) {
				printf("vexec_verify: Attempt to "
				    "execute %s directly: Enforcing "
				    "correct type. (uid=%u, pid=%u, "
				    "ppid=%u, pgrp=%u, dev=%d, "
				    "inode=%u)\n", name,
				    p->p_ucred->cr_uid, p->p_pid,
				    p->p_pptr->p_pid, p->p_pgrp->pg_id,
				    va->va_fsid, 
				    va->va_fileid);

				return (EPERM);
			} else {
				return (0);
			}
		}

		/* Calculate fingerprint for the inode. */
		error = vexec_fp_calc(p, vp, vhe, va->va_size, digest);
		if (error) {
			vexec_dprintf(("vexec_verify: calculation error\n"));
			return (error);
		}

		/*
		 * We don't need to check the fingerprint type here
		 * because vexec_fp_calc() does that too, and if it failed
		 * there we would already be error'd out.
		 */
		if (vexec_fp_cmp(vhe, digest) == 0) {
			if (vhe->vhe_type == VEXEC_INDIRECT) {
				vp->v_fpstatus = FP_INDIRECT;
			} else {
				vp->v_fpstatus = FP_VALID;
			}
		} else {
			vp->v_fpstatus = FP_NOMATCH;
		}
	}

out:
        switch (vp->v_fpstatus) {
	case FP_NOTEVAL:
		/* Should not happen. */
		panic("vexec_verify: Not-evaluated status post-evaluation."
		    " Inconsistency detected. Report a bug.");

	case FP_VALID:
		/* Valid fingerprint. */
		if (security_vexec && security_vexec_verbose)
			printf("vexec_verify: Fingerprint matches. "
			    "(file=%s, inode=%u, dev=%d)\n", name,
			    va->va_fileid, va->va_fsid);
		break;

	case FP_INDIRECT:
		/* Fingerprint is okay; Make sure it's indirect execution. */
		if (flag == VEXEC_DIRECT) {
			/*
			 * Notify about direct execution attempts of
			 * programs marked "indirect only".
			 */
			printf("vexec_verify: Attempt to execute %s "
			    "(dev=%d, inode=%u) directly by uid=%u "
			    "(pid=%u, ppid=%u, gppid=%u)\n", name,
			    va->va_fsid, va->va_fileid,
			    p->p_ucred->cr_uid, p->p_pid, 
			    p->p_pptr->p_pid, p->p_pptr->p_pptr->p_pid);

			error = EPERM;
		}

		break;

	case FP_NOMATCH:
		/* Fingerprint mismatch. Deny execution. */
		if (security_vexec) {
			printf("vexec_verify: Fingerprint mismatch "
			    "for %s (dev=%d, inode=%u). Execution "
			    "attempt by uid=%u, pid=%u.\n", name,
			    va->va_fsid, va->va_fileid, 
			    p->p_ucred->cr_uid, p->p_pid);

			error = EPERM;
		}

		break;

	case FP_NOENTRY:
		/* No entry in the list. */
		if (security_vexec) {
			if (security_vexec_verbose)
				printf("vexec_verify: No fingerprint for %s "
				    "(dev=%d, inode=%u)\n", name, va->va_fsid,
				    va->va_fileid);

			if ((securelevel >= VEXEC_SECURELEVEL) &&
			    (security_vexec_strict))
				error = EPERM;
		}

		break;

	default:
		/*
		 * Should never happen.
		 * XXX: Print vnode/process?
		 */
		panic("vexec_verify: Not-evaluated status post-evaluation."
		    " Report a bug. (vnode=%p, pid=%u)", vp, p->p_pid);
        }

	return (error);
}

/*
 * Operations needed to be done before read()/write() requests to
 * (fingerprinted) vnodes.
 */
int vexec_openchk(struct proc *p, struct vnode *vp, char *pathbuf) {
	struct vexec_hash_entry *vhe;
	struct vattr va;
	u_char fp[FPLEN_MAX];
	int error;

	error = VOP_GETATTR(vp, &va, p->p_ucred, p);
	if (error)
		return (error);

	switch (vp->v_fpstatus) {
	case FP_VALID:
	case FP_INDIRECT:
		return (0);

	case FP_NOMATCH:
		if (security_vexec_verbose) {
			printf("vexec_openchk: Denying access to "
			    "%s: Bad fingerprint. (uid=%u, "
			    "pid=%u, inode=%u, dev=%d)\n",
			    pathbuf, p->p_ucred->cr_uid, p->p_pid,
			    va.va_fileid, va.va_fsid);
		}

		return (EPERM);
	}

	if (vp->v_fpstatus == FP_NOENTRY) {
		if ((securelevel >= VEXEC_SECURELEVEL) &&
		    (security_vexec_strict)) {
			printf("vexec_openchk: Denying access to %s: "
			    "System in highly-secure mode, strict. "
			    "(inode=%u, dev=%d, status=%u)\n",
			    va.va_fileid, va.va_fsid, vp->v_fpstatus);

			return (EPERM);
		}

		vexec_dprintf(("vexec_openchk: Allowing access to unknown"
		    " file. (file=%s, inode=%u, dev=%d, code=%d)\n",
		    pathbuf, va.va_fileid, va.va_fsid, vp->v_fpstatus));

		return (0);
	}

	vhe = vexec_lookup(va.va_fsid, va.va_fileid);

	/*
	 * If we got an entry for the device, but not for the file
	 * itself, set the fingerprint status to "no entry", tell the
	 * user, and if securelevel/strict mode says so -- deny access.
	 */
	if (vhe == NULL) {
		vp->v_fpstatus = FP_NOENTRY;

		vexec_dprintf(("vexec_openchk: No entry for %s. (dev=%d, "
		    "inode=%u)\n", pathbuf, va.va_fsid, va.va_fileid));

		if ((securelevel >= VEXEC_SECURELEVEL) &&
		    (security_vexec_strict)) {
			if (security_vexec_verbose)
				printf("vexec_openchk: Denying access to "
				    "%s: No fingerprint; System "
				    "highly-secure, strict. (uid=%u, "
				    "pid=%u, inode=%u, dev=%d)\n",
				    pathbuf, p->p_ucred->cr_uid, p->p_pid,
				    va.va_fileid, va.va_fsid);

			return (EPERM);
		} else
			return (error);
	}

	/* Calculate fingerprint. */
	error = vexec_fp_calc(p, vp, vhe, va.va_size, fp);
	if (error)
		return (error);

	/* Compare fingerprints. */
	if ((vexec_fp_cmp(vhe, fp) == 0)) {
		/* Fingerprint ok. */
		vp->v_fpstatus = FP_VALID;

		if (security_vexec_verbose)
			printf("vexec_openchk: Fingerprint matches. "
			    "(file=%s, dev=%d, inode=%u)\n", pathbuf,
			    va.va_fsid, va.va_fileid);
	} else {
		/* Fingerprint mismatch. */
		vp->v_fpstatus = FP_NOMATCH;

		printf("vexec_openchk: Fingerprint mismatch. "
		    "(file=%s, dev=%d, inode=%u)\n", pathbuf, va.va_fsid,
		    va.va_fileid);

		error = EPERM;
	}

	return (error);
}

/*
 * Vexec remove policy code. If we have an entry for the file in our
 * tables, we disallow removing if the securelevel is high.
 */
int vexec_removechk(struct proc *p, struct vnode *vp, char *pathbuf) {
	struct vexec_hashtbl *tbl;
	struct vexec_hash_entry *vhe;
	struct vattr va;
	const char *errmsg;
	int error;

	error = VOP_GETATTR(vp, &va, p->p_ucred, p);
	if (error)
		return (error);

	switch (vp->v_fpstatus) {
	case FP_VALID:
	case FP_INDIRECT:
	case FP_NOMATCH:
		/*
		 * These can be set only for files we have an entry for
		 * and hence for evaluated the fingerprint. Remove if
		 * securelevel allows.
		 */
		if ((securelevel >= VEXEC_SECURELEVEL) ||
		    security_vexec_strict) {
			if (!security_vexec_strict)
				errmsg = "strict mode";
			else
				errmsg = "system highly-secure";

			/* No way. */
			printf("vexec_removechk: Denying unlink request "
			    "for %s from uid=%u: File in tables, %s. "
			    "(pid=%u, inode=%u, dev=%d)\n", pathbuf,
			    p->p_ucred->cr_uid, errmsg, p->p_pid,
			    va.va_fileid, va.va_fsid);

			error = EPERM;
		} else {
			vhe = vexec_lookup(va.va_fsid, va.va_fileid);
			if (vhe == NULL) {
				panic("vexec_removechk: Inconsistency! "
				    "File has post-evaluation status, "
				    "but not in lists. Report a bug.");
			}

			if (security_vexec_verbose) {
				printf("vexec_removechk: Removing entry "
				    "from Vexec table. (file=%s, "
				    "inode=%u, dev=%d)\n", pathbuf,
				    va.va_fileid, va.va_fsid);
			}

			goto vexec_rm;
		}

		break;

	case FP_NOENTRY:
		/* Evaluated - no entry. */
		break;

	case FP_NOTEVAL:
		/*
		 * Could be we don't have an entry for this, but we can't
		 * risk an unevaluated file or vnode cache flush.
		 */
		vhe = vexec_lookup(va.va_fsid, va.va_fileid);
		if (vhe == NULL) {
			break;
		}

		if ((securelevel >= VEXEC_SECURELEVEL) ||
		    security_vexec_strict) {
			if (!security_vexec_strict)
				errmsg = "strict mode";
			else
				errmsg = "system highly-secure";

			printf("vexec_removechk: Denying unlink request "
			    "for %s from uid=%u: File in tables, %s. "
			    "(pid=%u, inode=%u, dev=%d)\n", pathbuf,
			    p->p_ucred->cr_uid, errmsg, p->p_pid,
			    va.va_fileid, va.va_fsid);

			error = EPERM;
		} else {
			goto vexec_rm;
		}

		break;
	}

	return (error);

vexec_rm:
	tbl = vexec_tblfind(va.va_fsid);
	if (tbl == NULL) {
		panic("vexec_removechk: Inconsistency: Could not get "
		    "table for file in lists. Report a bug.");
	}

	tbl->hash_size--;
	LIST_REMOVE(vhe, vhe_list);
	free(vhe, M_TEMP);

	return (error);
}
