#!/bin/sh
#
# Copyright (c) 2004 br1an <brian@ethernet.org>
#
#   Stephanie for OpenBSD 3.6
#   http://ethernet.org/~brian/Stephanie/
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#
# This script will scan (ffs) filesystems for files and will add their
# digital fingerprint to a database.
#
# Options:
#   -d <dir>	start scanning from <dir>
#   -r		do recursive scanning
#   -a		scan all files (not only executables)
#   -q		be quiet (no output)
#
# Usage pattern: fpgen.sh [flags] <hash> <database file>
# Examples:
#   ./fpgen.sh -r md5 vexec.conf (scan recursively from current dir)
#   ./fpgen.sh -ra sha1 vexec.conf (all files, recursively from curent dir)
#   ./fpgen.sh -d /bin rmd160 vexec.conf (only executables in /bin)

# Defaults
find_pattern="-perm -0100 -or -perm -0010 -or -perm -0001"
recurse="-maxdepth 1"
rootdir="$PWD"
howdeep="non-recursive"
scanwhat="executables only"
quiet=0

# Sanity
if [ $# -lt 2 -o $# -gt 5 ]; then
	echo "Usage: $0 [-a|-r|-d <dir>] [-q] <hash_type> <fingerprint_file>"
	echo "(a)ll files; (r)ecursive scan; (d)irectory root."
	exit 2
fi

# Parse arguments: Build search pattern and set root dir.
set -- `getopt qard: $*`
for i do
	case "$i" in
		-a) scanwhat="all files"; find_pattern="! -perm 0000"; shift;;
		-d) rootdir=$2; shift; shift;;
		-r) howdeep="recursive"; recurse=; shift;;
		-q) quiet=1; shift;;
	esac
done

# Set hash type and command (ghetto) and output file
hashcmd=$2;
dbfile=$3;

# Make sure we have a hash type
if [ -z $hashcmd ]; then
	echo "No hash! quitting."
	exit;
fi

# Make sure we have a target file.
if [ -z $dbfile ]; then
	echo "No dbfile! quitting."
	exit
fi

if [ $quiet == 0 ]; then
	echo "Executing \"$hashcmd\" on \"$rootdir\" ($howdeep), storing in \"$dbfile\"."
	echo "Scanning for $scanwhat."
fi

# Backup existing file (if got one)
if [ -e $dbfile ]; then
	if [ $quiet == 0 ]; then
		echo "Backing up existing fingerprint file to \"$dbfile.old\""
	fi

	mv $dbfile $dbfile.old
fi

find $rootdir $recurse \( \( -fstype ffs -and -type f -and \( $find_pattern \) \) \
-or -name "lib*so*" \) | egrep -v '^/proc.*' | \
while read file; do
	if [ $quiet == 0 ]; then
		echo "Fingerprinting $file"
	fi

	hash=`$hashcmd $file | cut -d'=' -f2`
	flag=""

	# Non-executables need the "FILE" flag
	if [ ! -x $file ]; then
		flag=" FILE"
	fi

	# Add it to the database
	if [ ${hash} ]; then
		echo "$file $hashcmd ${hash}${flag}" >> $dbfile
	fi
done

if [ $quiet == 0 ]; then
	echo "\n##################################################################"
	echo "\n    PLEASE VERIFY CONTENTS OF $dbfile AND SET THE \"INDIRECT\"\t"
	echo "    FLAG WHERE APPROPRIATE (HINT: INTERPRETERS SUCH AS PERL).\t\n"
	echo "##################################################################"
fi
