/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _SYS_SECURITY_H_
#define _SYS_SECURITY_H_

/*
 * Stephanie security features.
 */
#define	KERN_SECURITY_TRUSTED		1	/* bool: curproc trusted? */
#define	KERN_SECURITY_TRUSTGID		2	/* int64_tt: trust group id */
#define	KERN_SECURITY_TRUST_OVERRIDE	3	/* bool: trust override */
#define KERN_SECURITY_TPE		4	/* bool: trusted path exec */
#define KERN_SECURITY_VEXEC		5	/* node: verified exec */
#define KERN_SECURITY_PRIVACY		6	/* bool: process privacy */
#define KERN_SECURITY_MAXID		7	/* int: max */

#define CTL_KERN_SECURITY_NAMES { \
	{ 0, 0 }, \
	{ "trusted", CTLTYPE_INT }, \
	{ "trust_gid", CTLTYPE_QUAD }, \
	{ "trust_override", CTLTYPE_INT }, \
	{ "tpe", CTLTYPE_INT }, \
	{ "vexec", CTLTYPE_NODE }, \
	{ "privacy", CTLTYPE_NODE }, \
}

/* Bit offsets */
#define KERN_PRIVACY_OP		1
#define KERN_PRIVACY_PROCFS	2
#define	KERN_PRIVACY_NETSTAT	3
#define	KERN_PRIVACY_W		4
#define	KERN_PRIVACY_WHO	5
#define	KERN_PRIVACY_LAST	6
#define	KERN_PRIVACY_FINGER	7
#define KERN_PRIVACY_MAXID	8

#define CTL_KERN_PRIVACY_NAMES { \
	{ 0, 0 }, \
	{ "op", CTLTYPE_INT }, \
	{ "procfs", CTLTYPE_INT }, \
	{ "netstat", CTLTYPE_INT }, \
	{ "w", CTLTYPE_INT }, \
	{ "who", CTLTYPE_INT }, \
	{ "last", CTLTYPE_INT }, \
	{ "finger", CTLTYPE_INT }, \
}

#ifdef _KERNEL
extern int64_t security_trust_gid;
extern int security_trust_override;
extern int security_tpe;
extern int security_privacy;
extern int security_privacy_procfs;

/*
 * Trusted path checking macro.
 *
 * A path is considered "trusted" if it's owned by the superuser, and
 * its writable neither for 'group' or 'other'.
 */
#define	TRUSTED_PATH(va)	(((va).va_uid == 0) && \
				 !((va).va_mode & (S_IWGRP | S_IWOTH)))

#define	TRUST_INVAL		(-1)

#define	BITMAP_SET(map, bit)	((map) |= (1 << (bit)))
#define	BITMAP_ISSET(map, bit)	((map) & (1 << (bit)))
#define	BITMAP_CLR(map, bit)	((map) &= ~(1 << (bit)))
#define	BITMAP_ISCLR(map, bit)	(!(BITMAP_ISSET(map, bit)))

/*
 * Function prototypes.
 */
/* kern/kern_security.c: */
int trusted(struct proc *);
int private(struct proc *, int);
int proc_cansee(struct proc *, struct proc *);
int sysctl_security(int *, u_int, void *, size_t *, void *, size_t, struct proc *);

#else

#ifdef _SYS_SYSCTL_H_
int security_check(int, int);

/*
 * Check if a certain feature in the security node is enabled. Placing it
 * here to not mess with libc and adding files etc. Much simpler.
 */
int security_check(int third, int fourth) {
	int mib[4], status, r;
	size_t len = 4, size = sizeof(status);

	mib[0] = CTL_KERN;
	mib[1] = KERN_SECURITY;
	mib[2] = third;
	mib[3] = fourth;

	r = sysctl(mib, len, &status, &size, NULL, 0);
	if (r < 0) {
		return (1);
	}

	return (status ? 0 : -1);
}

/* Useful macros. */
#define	TRUSTED()		(security_check(KERN_SECURITY_TRUSTED, 0) == 0)
#define TRUST_OVERRIDE()	(security_check(KERN_SECURITY_TRUST_OVERRIDE, 0) == 0)
#define	PRIV_NETSTAT()		(security_check(KERN_SECURITY_PRIVACY, KERN_PRIVACY_NETSTAT) == 0)
#define	PRIV_W()		(security_check(KERN_SECURITY_PRIVACY, KERN_PRIVACY_W) == 0)
#define	PRIV_WHO()		(security_check(KERN_SECURITY_PRIVACY, KERN_PRIVACY_WHO) == 0)
#define	PRIV_LAST()		(security_check(KERN_SECURITY_PRIVACY, KERN_PRIVACY_LAST) == 0)
#define	PRIV_FINGER()		(security_check(KERN_SECURITY_PRIVACY, KERN_PRIVACY_FINGER) == 0)

#endif /* _SYS_SYSCTL_H_ */

#endif /* _KERNEL */

#endif /* _SYS_SECURITY_H_ */
