/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef _SYS_VEXEC_H_
#define _SYS_VEXEC_H_

#include <sys/types.h>
#include <sys/device.h>
#include <sys/hash.h>
#include <crypto/md5.h>
#include <crypto/sha1.h>
#include <crypto/sha2.h>
#include <crypto/rmd160.h>
#include <sys/queue.h>

/* These serve as both types and indexes into the vexec hashes table. */
#define	FPTYPE_MD5	0
#define	FPTYPE_SHA1	1
#define	FPTYPE_SHA256	2
#define	FPTYPE_SHA384	3
#define	FPTYPE_SHA512	4
#define	FPTYPE_RMD160	5
#define FPTYPE_MAXID	6

/* Fingerprint lengths. Make sure FPLEN_MAX is okay. */
#define	FPLEN_MD5	MD5_DIGEST_LENGTH
#define	FPLEN_SHA1	SHA1_DIGEST_LENGTH
#define	FPLEN_SHA256	SHA256_DIGEST_LENGTH
#define	FPLEN_SHA384	SHA384_DIGEST_LENGTH
#define	FPLEN_SHA512	SHA512_DIGEST_LENGTH
#define	FPLEN_RMD160	20
#define FPLEN_MAX	64

/*
 * Entries in the Vexec lists can have either of these types:
 *  - VEXEC_INDIRECT: Only indirect (via scripts) execution allowed;
 *    arguments which can resolve to pathnames are checked for trust if
 *    TPE is enabled.
 *  - VEXEC_DIRECT: Both direct (via execve()) and indirect execution
 *    is allowed. Arguments are not checked for trust.
 *  - VEXEC_SHELL: Allows direct/indirect execution; arguments are checked
 *    for trust.
 *  - VEXEC_FILE: Maintain integrity during open() calls. (for read/write).
 *
 * The Vexec direct/indirect flags are also used *outside* of Vexec to
 * solve the "TPE bypass problem". Up to Stephanie for 3.2, you
 * had to set immutable flag on interpreters to prevent them from
 * bypassing the TPE restrictions.
 *
 * This had several implications, including inability to "su" to
 * to an untrusted user (which was solved with another ugly hack..)
 * and the need to bring the system to single-user to update an
 * interpreter/shell.
 *
 * As of Stephanie for OpenBSD 3.4, you can use the direct/indirect flags
 * on executables (via vexec.conf) to make interpreters load only 
 * indirectly (from scripts), and combined with TPE, it solves the problem
 * of users feeding an interpreter from stdin (fd 0), both solving the
 * "su" problem and removing the ugly code from the read() and dup2()
 * syscalls and the extra process flag.
 *
 * In a later revision of Stephanie for OpenBSD 3.6, a new type of
 * Vexec entry was added - VEXEC_SHELL - to allow several programs who
 * may act as interpreters have their arguments checked without forcing
 * them to be executed indirectly. This usually applies to shells that
 * until now there was no easy solution to maintain trust of shell
 * scripts ran. (sh /path/to/script vs. /bin/sh on login)
 */
#define VEXEC_DIRECT		0
#define VEXEC_INDIRECT		1
#define VEXEC_SHELL		2
#define VEXEC_FILE		3

/*
 * Node data for sysctl (pulled from kern/kern_security.c and userland
 * sysctl, referenced from sys/security.h).
 */
#define KERN_VEXEC_OP		1
#define KERN_VEXEC_VERBOSE	2
#define	KERN_VEXEC_STRICT	3
#define KERN_VEXEC_MAXID	4

#define	CTL_KERN_VEXEC_NAMES { \
	{ 0, 0 }, \
	{ "op", CTLTYPE_INT }, \
	{ "verbose", CTLTYPE_INT }, \
	{ "strict", CTLTYPE_INT }, \
}

/*
 * VEXEC_PRELOAD is used to prepare the kernel memory required for the
 * management of the Vexec hash tables. It contains a parameter with
 * information about a single device and the amount of fingerprinted files
 * on it.
 *
 * VEXEC_LOAD is used to load a fingerprint to the kernel.
 */
#define	VEXEC_PRELOAD	_IOW('S', 0x1, struct vexec_hash_param)
#define VEXEC_LOAD	_IOW('S', 0x2, struct vexec_params)

/* Parameters for vexectioctl() using VEXEC_PRELOAD. */
struct vexec_hash_param {
	dev_t hp_dev;		/* Device ID. */
	size_t hp_size;		/* Number of items on that device. */
};

/* Parameters for vexecioctl() using VEXEC_LOAD. */
struct vexec_params {
	u_char vxp_type;		/* Entry type. */
	u_char vxp_fptype;		/* Hash type. */
	u_char vxp_file[MAXPATHLEN];	/* Filename. */
	u_char vxp_fp[FPLEN_MAX];	/* Fingerprint. */
};

#ifdef _KERNEL
/*
 * Vexec supports several types of digital-checksumming hashes, and
 * allows you to easily add new types if available.
 *
 * Vexec uses a table describing available hashing types and their
 * associated data (checksum length, hashing functions..) that's
 * available only in the kernel's address space; but since we need
 * type/length indication in userland aswell (for parsing the config
 * file), we maintain a list of types/length available in userland
 * too.
 *
 * To add more hash types:
 *
 *  - Make sure you have hashing routines that can be interfaced by 
 *    the function types (see below).
 *  - Add entries for the new type (type/length) below. Re-adjust the
 *    maximum fingerprint length if needed.
 *  - Add an entry for the new hash type in kern/kern_vexec.c.
 *  - Change the userland control utility ("vexecctl", parse.y) to
 *    recognize the new type.
 */

/* Function types: init, update, final. */
typedef void (*VEXEC_INIT_FN)(void *);
typedef void (*VEXEC_UPDATE_FN)(void *, u_char *, u_int);
typedef void (*VEXEC_FINAL_FN)(u_char *, void *);

/* Structure describing a hash type. */
struct vexec_hash {
	u_char vxh_type;		/* Hash type. */
	size_t vxh_len;			/* Hash length. */
	size_t vxh_ctxsize;		/* Calculation context size. */
	VEXEC_INIT_FN vxh_init;		/* Context "init" function. */
	VEXEC_UPDATE_FN vxh_update;	/* Context "update" function. */
	VEXEC_FINAL_FN vxh_final;	/* Context "final" function. */
};

/* The table (set in kern/kern_vexec.c). */
extern struct vexec_hash vexec_hash_list[];

/* We need this for "diagnostics". */
#define FPTYPE_OK(type)	((type) < (u_char) FPTYPE_MAXID)

/* An entry in the per-device hash table. */
struct vexec_hash_entry {
	ino_t vhe_inode;				/* Inode number. */
	unsigned char vhe_type;				/* Entry type. */
	unsigned char vhe_fptype;			/* Hash type. */
	unsigned char vhe_fp[FPLEN_MAX];		/* Fingerprint. */
	LIST_ENTRY(vexec_hash_entry) vhe_list;		/* List pointer. */
};

LIST_HEAD(vexec_hashhead, vexec_hash_entry) *hash_tbl;

/* Vexec hash table information. */
struct vexec_hashtbl {
	struct vexec_hashhead *hash_tbl;
	size_t hash_size;	/* Number of slots in the table. */
	dev_t hash_dev;		/* Device ID the hash table refers to. */
	LIST_ENTRY(vexec_hashtbl) hash_list;
};

/* Global list of hash tables. */
LIST_HEAD(, vexec_hashtbl) vexec_tables;

/*
 * This makes it easier to add hash functions that use the default generic
 * naming scheme. (HASHInit, HASHUpdate, HASHFinal, FPTYPE_HASH,
 * FPLEN_HASH and HASH_CTX)
 *
 * VEXEC_HASHINIT2() has been added to support conventions like HASH_Init
 * etc., used in the SHA256, SHA384, and SHA512 routines.
 */
#define	VEXEC_HASHINIT(hash)	FPTYPE_ ## hash, \
				FPLEN_ ## hash, \
				sizeof(hash ## _CTX), \
				(VEXEC_INIT_FN) hash ## Init, \
				(VEXEC_UPDATE_FN) hash ## Update, \
				(VEXEC_FINAL_FN) hash ## Final

#define	VEXEC_HASHINIT2(hash)	FPTYPE_ ## hash, \
				FPLEN_ ## hash, \
				sizeof(hash ## _CTX), \
				(VEXEC_INIT_FN) hash ## _Init, \
				(VEXEC_UPDATE_FN) hash ## _Update, \
				(VEXEC_FINAL_FN) hash ## _Final

/* Compute hash mask. */
#define	VEXEC_HASH_MASK(tbl)	((tbl)->hash_size - 1)

/* Returns an index. */
#define	VEXEC_HASH(tbl, inode)	\
	(hash32_buf(&(inode), sizeof((inode)), HASHINIT) & VEXEC_HASH_MASK(tbl))

/*
 * To comply with the "last line of defense" features of securelevel,
 * Vexec allows you to specify at what securelevel it will stop being
 * nice and will start treating warnings as errors. The implications,
 * as far as Vexec is concerned, of the securelevel set here are:
 *
 *  - Executables/files with no entries will be denied access to.
 *  - Write requests to files with vaild fingerprints will be denied.
 *
 * Basically, it's recommended that you don't change this setting.
 * However, if you want to change it, make sure you have a fingerprint
 * for every file on the system you intend to use. Note that any new
 * files you'll create after having the system at this securelevel
 * will need their fingerprints to be loaded...
 */
#define	VEXEC_SECURELEVEL	2

extern int security_vexec;
extern int security_vexec_verbose;
extern int security_vexec_strict;

/* Callback for hash traversal. */
typedef void (*VEXEC_CALLBACK)(struct vexec_hash_entry *, dev_t);

/*
 * Function prototypes.
 */
/* dev/vexecx.c: */
void vexecattach(struct device *, struct device *, void *);
int vexecopen(dev_t, int, int, struct proc *);
int vexecclose(dev_t, int, int, struct proc *);
int vexecioctl(dev_t, u_long, caddr_t, int, struct proc *);
/* int vexec_inodeadd(struct vexec_dev_list *, struct proc *, struct vnode *,    struct vexec_params *, struct vattr); */
/* struct vexec_dev_list *vexec_finddev(dev_t, struct vexec_devhead *); */

/* kern/kern_vexev.c: */
/* void vexec_fp_convert(u_char *, u_char *, size_t); */
int vexec_fp_calc(struct proc *, struct vnode *,
    struct vexec_hash_entry *, u_quad_t, u_char *);
int vexec_fp_cmp(struct vexec_hash_entry *, u_char *);

struct vexec_hashtbl *vexec_tblfind(dev_t);
struct vexec_hash_entry *vexec_lookup(dev_t, ino_t);
int vexec_hashadd(struct vexec_hashtbl *, struct vexec_hash_entry *);
void vexec_hashprint(struct vexec_hash_entry *, dev_t);
int vexec_tblwalk(VEXEC_CALLBACK *);

int vexec_verify(struct proc *, struct vnode *, struct vattr *, u_char *, int);
int vexec_openchk(struct proc *, struct vnode *, char *);
int vexec_removechk(struct proc *, struct vnode *, char *);
#endif /* _KERNEL */

/* Use for debugging only. */
#ifdef VEXEC_DEBUG
#define vexec_dprintf(x) printf x
#else
#define vexec_dprintf(x)
#endif /* VEXEC_DEBUG */

#endif /* _SYS_VEXEC_H_ */
