%{
/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <sys/vexec.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>

#include "y.tab.h"
#include "vexecctl.h"

/*
 * Max nesting depth. It's 16 both because it's a nice number
 * and it's the age I first made my Stephanie own release. :)
 */
#define MAX_INCL_DEPTH	16

YY_BUFFER_STATE incl_stack[MAX_INCL_DEPTH];
int incl_depth = 0;

struct my_stack *filestack;
u_int64_t entries;
%}

INCLUDE		[Ii][Nn][Cc][Ll][Uu][Dd][Ee]
STRING		[0-9a-zA-Z]+

%s incl

%%

{INCLUDE}	{ BEGIN(incl); }

<incl>{STRING}/([ \t]*\n) {
			/*
			 * Make sure we're not going too deep...
			 * XXX: Should we skip the include and continue
			 *      with the rest of the file, or error?
			 */
			if (incl_depth >= MAX_INCL_DEPTH) {
				(void) fprintf(stderr, "ERROR: "
				    "Nesting too deep: File %s "
				    "included from %s.\n", yytext,
				    (filestack ? filestack->name :
				    "(unknown)"));

				filestack_free(filestack);

				exit(1);
			}

			/*
			 * Save current lexer state.
			 */
			incl_stack[incl_depth++] = YY_CURRENT_BUFFER;

			/*
			 * Phase 1: Open the file.
			 * Phase 2: Already opened, do nothing.
			 */
			if (phase == 2)
				goto already_opened;

			yyin = openlock(yytext);
			if (!yyin) {
				err(1, "Can't open %s", yytext);
			}			

already_opened:
			/*
			 * Let the user know, if verbose...
			 */
			if (verbose) {
				(void) fprintf(stderr, "=> Including "
				    "\"%s\"\n", yytext);
			}

			if (!no_mem && filestack) {
				filestack->next = (struct my_stack *)
				    malloc(sizeof(*(filestack->next)));
				if (!(filestack->next)) {
					warnx("WARNING: Memory shortage"
					    " during runtime. Turning"
					    " on verbose mode.");
					filestack_free(filestack);
					no_mem = 1;
					verbose = 1;
				} else {
					filestack->next->prev = filestack;
					filestack = filestack->next;
					filestack->name = strdup(yytext);
					if (!(filestack->name)) {
						warnx("WARNING: Memory"
						    " shortage during "
						    "runtime. Turning "
						    "on verbose mode.");
						filestack_free(filestack);
						no_mem = 1;
						verbose = 1;
					}

					filestack->line = 1;
				}
			}

			/*
			 * Create a new lexer state buffer, and switch
			 * to it.
			 */
			yy_switch_to_buffer(yy_create_buffer(yyin,
			    YY_BUF_SIZE));

			/*
			 * Return to initial start condition, and
			 * start parsing the new file.
			 */
			BEGIN(0);
		}

<incl>[ \t]+	;

<incl>.		{
			yyerror("Unexpected token in include statement");

			/*
			 * Eat up current line and continue parsing.
			 * XXX: Make -c (checking mode) and only continue
			 *      when checking the files, but when
			 *      actually loading, error out?
			 */
			while (input() != '\n');

			BEGIN(0);
		}

<<EOF>>		{
			if (verbose) {
				if (no_mem) {
					(void) fprintf(stderr, "=> EOF\n");
				} else {
					(void) fprintf(stderr, "=> EOF"
					    " from \"%s\"\n", filestack->name);
				}
			}

			if (--incl_depth < 0) {
				yyterminate();
			} else {
				if (!no_mem && filestack) {
					/*
					 * Phase 1: Rewind file.
					 * Phase 2: Close file.
					 */
					if (phase == 1)
						lseek(__sfileno(yyin), 0L, SEEK_SET);
					else
						fclose(yyin);

					filestack = filestack->prev;
					free(filestack->next);
				}

				yy_delete_buffer(YY_CURRENT_BUFFER);
				yy_switch_to_buffer(incl_stack[incl_depth]);
			}
		}

	/* path */
\/[^ \t]+	{
			yylval.string = strdup(yytext);
			return (PATH);
		}

	/* string (fingerprint, type, options) */
{STRING} 	{
			yylval.string = strdup(yytext);
			if (!(yylval.string)) {
				(void) fprintf(stderr, "WARNING: Out"
				    " of memory during run-time! "
				    "Attempting recover...\n");

				/*
				 * Sleep a little bit and let the
				 * machine calm down. ;)
				 */
				usleep(600);

				yylval.string = strdup(yytext);
				if (!(yylval.string)) {
					(void) fprintf(stderr, "ERROR:"
					    " No memory. Aborting.\n");
					exit(1);
				}

				(void) fprintf(stderr, "Recovered.\n");
			}

			return (STRING);
		}


	/* comments, white-outs */
[ \t\r]		|
#.*		;
#.*\n		|
^\n		{ (filestack->line)++; }

	/* eol on a line with data. need a call to ioctl, return eol */
\n		{
			(filestack->line)++;
			return (EOL);
		}

.		{ yyerror("Invalid character"); }

%%

int yywrap(void) {
	return (1);
}

void yyerror(const char *string) {
	if (!no_mem && filestack) {
		(void) fprintf(stderr, "%s:%u: %s at \"%s\"\n",
		    filestack->name, filestack->line, string, yytext);
	} else {
		(void) fprintf(stderr, "%s at \"%s\"\n", string, yytext);
	}
}
