%{
/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * $NetBSD: veriexecctl_parse.y,v 1.1 2002/11/23 10:52:49 blymn Exp $
 */

#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/mount.h>

#include <sys/vexec.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <err.h>

#include "vexecctl.h"

struct vexec_params params;

struct vexec_hashmap {
	u_int hm_type;
	char *hm_name;
	size_t hm_len;
};

struct vexec_hashmap vexec_hashmaps[] = {
    { FPTYPE_MD5, "md5", FPLEN_MD5 },
    { FPTYPE_SHA1, "sha1", FPLEN_SHA1 },
    { FPTYPE_SHA256, "sha256", FPLEN_SHA256 },
    { FPTYPE_SHA384, "sha384", FPLEN_SHA384 },
    { FPTYPE_SHA512, "sha512", FPLEN_SHA512 },
    { FPTYPE_RMD160, "rmd160", FPLEN_RMD160 },
    { 0, NULL, 0 },
};
size_t vexec_hashmap_len = sizeof(vexec_hashmaps) / sizeof(struct vexec_hashmap);
u_int map_index;

%}

%union {
	char *string;
	int intval;
}

%token <string> PATH
%token <string> STRING
%token EOL

%%

statement	:	/* empty */
		|	statement path type fingerprint flags eol {
				struct stat sb;

				if (phase == 2) {
					phase2_load();
					goto phase_2_end;
				}

				if (stat(params.vxp_file, &sb) == 0) {
					struct vexec_up *p;

					/* Only regular files */
					if (!S_ISREG(sb.st_mode)) {
						(void) fprintf(stderr, "Line %u: "
						    "%s is not a regular file.\n",
						    filestack->line, params.vxp_file);
					}

					if ((p = dev_lookup(sb.st_dev)) != NULL) {
						(p->vu_param.hp_size)++;
					} else {
						if (verbose) {
							struct statfs sf;

							statfs(params.vxp_file, &sf);

							(void) printf(
							    " => Adding device ID %d. (%s)\n",
							    sb.st_dev, sf.f_mntonname);
						}

						dev_add(sb.st_dev);
					}
				} else {
					(void) fprintf(stderr, "Line %u: Can't stat %s.\n",
					    filestack->line, params.vxp_file);
				}

phase_2_end:
				bzero(&params, sizeof(params));
			}
		|	statement eol
		|	statement error eol {
				yyerrok;
			}
		;

path		:	PATH {
				strlcpy(params.vxp_file, $1, MAXPATHLEN);
			}
		;

type		:	STRING {
				if (phase == 1)
					goto skip_type;

				map_index = 0;
				while (vexec_hashmaps[map_index].hm_name != NULL) {
					if (strcasecmp($1,
					    vexec_hashmaps[map_index].hm_name) == 0) {
						params.vxp_fptype = vexec_hashmaps[map_index].hm_type;
						goto skip_type;
					}

					map_index++;
				}

				yyerror("Bad fingerprint type");
				YYERROR;
skip_type:
			}
		;


fingerprint	:	STRING {
				u_int count;

				if (phase == 1)
					goto skip_fingerprint;

				count = vexec_hashmaps[map_index].hm_len;

				if (convert($1, count, params.vxp_fp) == -1) {
					yyerror("Bad fingerprint");
					YYERROR;
				}

skip_fingerprint:
			}
		;

flags		:	/* empty */ {
				if (phase == 2)
					params.vxp_type = VEXEC_DIRECT;
			}
		|	flags flag_spec
		;

flag_spec	:	STRING {
				if (phase == 1)
					goto skip_flagspec;

				if (strcasecmp($1, "direct") == 0) {
					params.vxp_type = VEXEC_DIRECT;
				} else if (strcasecmp($1, "indirect") == 0) {
					params.vxp_type = VEXEC_INDIRECT;
				} else if (strcasecmp($1, "shell") == 0) {
					params.vxp_type = VEXEC_SHELL;
				} else if (strcasecmp($1, "file") == 0) {
					params.vxp_type = VEXEC_FILE;
				} else {
					yyerror("Bad option");
					YYERROR;
				}

skip_flagspec:
			}
		;

eol		:	EOL
		;

%%
		
/*
 * Takes the hexadecimal string pointed to by "fp" and converts it to a 
 * "count" byte binary number which is stored in the array pointed to
 * by "out".  Returns -1 if the conversion fails.
 */
int convert(char *fp, u_int count, u_char *out) {
        int i, value, error;

	error = 0;

	for (i = 0; i < count; i++) {
		if ((fp[2*i] >= '0') && (fp[2*i] <= '9')) {
			value = 16 * (fp[2*i] - '0');
		} else if ((fp[2*i] >= 'a') && (fp[2*i] <= 'f')) {
			value = 16 * (10 + fp[2*i] - 'a');
		} else {
			error = -1;
			break;
		}

		if ((fp[2*i + 1] >= '0') && (fp[2*i + 1] <= '9')) {
			value += fp[2*i + 1] - '0';
		} else if ((fp[2*i + 1] >= 'a') && (fp[2*i + 1] <= 'f')) {
			value += fp[2*i + 1] - 'a' + 10;
		} else {
			error = -1;
			break;
		}

		out[i] = value;
	}

	return (error);
}
