/*
 * Copyright (c) 2004 br1an <brian@ethernet.org>
 *
 *   Stephanie for OpenBSD 3.6
 *   http://ethernet.org/~brian/Stephanie/
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <sys/ioctl.h>
#include <sys/param.h>
#include <sys/queue.h>
#include <sys/vexec.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <err.h>

#include "vexecctl.h"

#define	VEXEC_DEVICE	"/dev/vexec"

#define	usage(code)	errx(code, "Usage: %s [-v] <signature_file>", \
			    __progname)

int fd, verbose = 0, no_mem = 0, phase;

FILE *openlock(const char *path) {
	int fd;

	fd = open(path, O_RDONLY|O_EXLOCK, 0);
	if (fd < 0)
		return (NULL);

	return (fdopen(fd, "r"));
}

struct vexec_up *dev_lookup(dev_t d) {
	struct vexec_up *p;

	for (p = CIRCLEQ_FIRST(&params_list); p != CIRCLEQ_END(&params_list);
	    p = CIRCLEQ_NEXT(p, vu_list)) {
		if (p->vu_param.hp_dev == d)
			return (p);
	}

	return (NULL);
}

struct vexec_up *dev_add(dev_t d) {
	struct vexec_up *up;

	up = calloc(1, sizeof(*up));
	if (up == NULL)
		err(1, "No memory");

	up->vu_param.hp_dev = d;
	up->vu_param.hp_size = 1;

	CIRCLEQ_INSERT_TAIL(&params_list, up, vu_list);

	return (up);
}

/* Load all devices, get rid of the list. */
int phase1_preload(void) {
	extern int errno;

	if (verbose)
		printf("Phase 1: Doing preload:\n");

	while (CIRCLEQ_FIRST(&params_list) != CIRCLEQ_END(&params_list)) {
		struct vexec_up *vup;

		vup = CIRCLEQ_FIRST(&params_list);

		if (ioctl(fd, VEXEC_PRELOAD, &(vup->vu_param)) < 0) {
			(void) fprintf(stderr, "Error in phase 1: Can't "
			    "preload for device %d: %s.\n",
			    vup->vu_param.hp_dev, strerror(errno));

			return (-1);
		}

		if (verbose) {
			printf(" => Preload successful for device %d. "
			    "(%u entries)\n", vup->vu_param.hp_dev,
			    vup->vu_param.hp_size);
		}

		CIRCLEQ_REMOVE(&params_list, vup, vu_list);
		free(vup);
	}

	return (0);
}

/*
 * Load the fingerprint. Assumes that the fingerprint pseudo-device is
 * opened and the file handle is in fd.
 */
void phase2_load(void) {
	extern int errno, fd;
	extern struct vexec_params params;

	if (ioctl(fd, VEXEC_LOAD, &params) < 0) {
		if (!no_mem && filestack) {
			(void) fprintf(stderr, "%s:%u: %s: %s\n",
			    filestack->name, filestack->line,
			    params.vxp_file, strerror(errno));
		} else {
			(void) fprintf(stderr, "%s: %s\n",
			    params.vxp_file, strerror(errno));
		}
	}
}

int main(int argc, char **argv) {
	char *infile;
	int c;
	extern char *__progname;

	if ((argc < 2) || (argc > 3)) {
		usage(1);
	}

	while ((c = getopt(argc, argv, "v")) != -1) {
		switch (c) {
		case 'v':
			verbose = 1;
			break;

		default:
			usage(1);
		}
	}

	argc -= optind;
	argv += optind;

	infile = argv[0];

	filestack = (struct my_stack *) malloc(sizeof(*filestack));
	if (!filestack) {
			(void) fprintf(stderr, "WARNING: Can't allocate"
			    " memory for filename stack. Turning on "
			    "verbose mode.\n");

			verbose = 1;
			no_mem = 1;
	} else {
		bzero(filestack, sizeof(*filestack));

		filestack->name = strdup(infile);
		if (!(filestack->name)) {
			/*
			 * Don't attempt further allocations. Having
			 * memory every once in a while may require more
			 * code to deal with more situations that may
			 * introduce slowness and possible inconsistency.
			 */
			(void) fprintf(stderr, "WARNING: Not enough "
			    "memory for smooth operation. Turning on "
			    "verbose mode.");

			free(filestack);

			verbose = 1;
			no_mem = 1;
		} else {
			filestack->line = 1;
		}
	}

	CIRCLEQ_INIT(&params_list);

	fd = open(VEXEC_DEVICE, O_WRONLY, 0);
	if (fd == -1) {
		err(1, "Failed to open pseudo-device");
	}

	if ((yyin = openlock(infile)) == NULL) {
		err(1, "Failed to open %s", infile);
	}

	/*
	 * Phase 1: Scan all config files, creating the list of devices
	 *	    we have fingerprinted files on, and the amount of
	 *	    files per device. Lock all files to maintain sync.
	 */
	phase = 1;

	if (verbose) {
		(void) fprintf(stderr, "Phase 1: Building hash table information:\n");
		(void) fprintf(stderr, "=> Parsing \"%s\"\n", infile);
	}

	yyparse();

	if (phase1_preload() < 0)
		exit(1);

	/*
	 * Phase 2: After we have a circular queue containing all the
	 * 	    devices we care about and the sizes for the hash
	 *	    tables, do a rescan, this time actually loading the
	 *	    file data.
	 */
	rewind(yyin);
	phase = 2;
	if (verbose) {
		(void) fprintf(stderr, "Phase 2: Loading per-file fingerprints.\n");
		(void) fprintf(stderr, "=> Parsing \"%s\"\n", infile);
	}

	yyparse();

	(void) close(fd);
	(void) fclose(yyin);

	return (0);
}
